from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_HYBRID_DOCUMENT

from ..events import (
    event_parsing_document_file_content_deleted,
    event_parsing_document_file_finished, event_parsing_document_file_submitted
)
from ..models import DocumentFilePageContent
from ..permissions import (
    permission_document_file_content_view, permission_document_file_parse,
    permission_document_type_parsing_setup
)
from ..utils import get_document_file_content

from .literals import TEST_DOCUMENT_CONTENT
from .mixins import (
    DocumentFileContentTestMixin, DocumentFileContentToolsViewsTestMixin,
    DocumentFileContentViewTestMixin, DocumentTypeContentViewsTestMixin
)


class DocumentFileContentViewsTestCase(
    DocumentFileContentViewTestMixin, DocumentFileContentTestMixin,
    GenericDocumentViewTestCase
):
    def test_document_file_content_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_content_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_content_view_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self._clear_events()

        response = self._request_test_document_file_content_view()
        self.assertContains(
            response=response, text=TEST_DOCUMENT_CONTENT, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_content_view_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_content_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_content_delete_single_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_content_delete_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_content_delete_single_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self._clear_events()

        response = self._request_test_document_file_content_delete_single_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].target, self.test_document_file)
        self.assertEqual(
            events[0].verb, event_parsing_document_file_content_deleted.id
        )

    def test_trashed_document_file_content_delete_single_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_content_delete_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_content_delete_multiple_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_content_delete_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_content_delete_multiple_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self._clear_events()

        response = self._request_test_document_file_content_delete_multiple_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].target, self.test_document_file)
        self.assertEqual(
            events[0].verb, event_parsing_document_file_content_deleted.id
        )

    def test_trashed_document_file_content_delete_multiple_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_content_delete_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            DocumentFilePageContent.objects.filter(
                document_file_page=self.test_document_file.pages.first()
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentFilePageContentViewsTestCase(
    DocumentFileContentViewTestMixin, DocumentFileContentTestMixin,
    GenericDocumentViewTestCase
):
    def test_document_file_page_content_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_content_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_content_view_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_content_view()
        self.assertContains(
            response=response, text=TEST_DOCUMENT_CONTENT, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_content_view_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_content_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_error_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_parsing_error_list_view()

        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_error_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self._clear_events()

        response = self._request_test_document_file_parsing_error_list_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_parsing_error_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_parsing_error_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentFileContentParsingViewsTestCase(
    DocumentFileContentViewTestMixin, GenericDocumentViewTestCase
):
    _skip_file_descriptor_test = True

    def test_document_file_parsing_download_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_content_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_download_view_with_access(self):
        self.expected_content_types = ('application/octet-stream',)

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self._clear_events()

        response = self._request_test_document_file_content_download_view()
        self.assertEqual(response.status_code, 200)

        self.assert_download_response(
            response=response, content=(
                ''.join(
                    get_document_file_content(
                        document_file=self.test_document_file
                    )
                )
            ),
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_parsing_download_view_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_content_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_content_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_submit_single_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_submit_single_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].target, self.test_document_file)
        self.assertEqual(
            events[0].verb, event_parsing_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self.test_document)
        self.assertEqual(events[1].target, self.test_document_file)
        self.assertEqual(
            events[1].verb, event_parsing_document_file_finished.id
        )

    def test_trashed_document_file_parsing_single_submit_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_submit_multiple_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_parsing_submit_multiple_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].target, self.test_document_file)
        self.assertEqual(
            events[0].verb, event_parsing_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self.test_document)
        self.assertEqual(events[1].target, self.test_document_file)
        self.assertEqual(
            events[1].verb, event_parsing_document_file_finished.id
        )

    def test_trashed_document_file_parsing_multiple_submit_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_parse
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_parsing_submit_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentTypeParsingViewsTestCase(
    DocumentFileContentToolsViewsTestMixin, GenericDocumentViewTestCase
):
    _skip_file_descriptor_test = True
    auto_upload_test_document = False

    # Ensure we use a PDF file
    test_document_filename = TEST_HYBRID_DOCUMENT

    def _get_document_file_content(self):
        return ''.join(
            list(
                get_document_file_content(
                    document_file=self.test_document_file
                )
            )
        )

    def test_document_parsing_error_list_view_no_permission(self):
        self._clear_events()

        response = self._request_document_parsing_error_list_view()
        self.assertEqual(response.status_code, 403)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_parsing_error_list_view_with_permission(self):
        self.grant_permission(permission=permission_document_file_parse)

        self._clear_events()

        response = self._request_document_parsing_error_list_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_view_no_permission(self):
        self._upload_test_document()

        self._clear_events()

        response = self._request_document_type_parsing_view()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_document_type.label
        )

        self.assertNotEqual(
            self._get_document_file_content(), TEST_DOCUMENT_CONTENT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_view_with_permission(self):
        self.grant_permission(permission=permission_document_file_parse)

        self._upload_test_document()

        self._clear_events()

        response = self._request_document_type_parsing_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._get_document_file_content(), TEST_DOCUMENT_CONTENT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].target, self.test_document_file)
        self.assertEqual(
            events[0].verb, event_parsing_document_file_submitted.id
        )

        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].action_object, self.test_document)
        self.assertEqual(events[1].target, self.test_document_file)
        self.assertEqual(
            events[1].verb, event_parsing_document_file_finished.id
        )


class DocumentTypeContentViewsTestCase(
    DocumentTypeContentViewsTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_document_type_parsing_settings_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_type_parsing_settings_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_view_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_parsing_setup
        )

        self._clear_events()

        response = self._request_test_document_type_parsing_settings_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.appearance.classes import Icon

icon_document_create_multiple = Icon(
    driver_name='fontawesome', symbol='upload'
)
icon_document_file_upload = Icon(
    driver_name='fontawesome', symbol='upload'
)

icon_source_backend_selection = Icon(
    driver_name='fontawesome-dual', primary_symbol='upload',
    secondary_symbol='plus'
)
icon_source_delete = Icon(driver_name='fontawesome', symbol='times')
icon_source_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_source_list = Icon(driver_name='fontawesome', symbol='upload')
icon_source_test = Icon(driver_name='fontawesome', symbol='check')

icon_upload_view_link = Icon(driver_name='fontawesome', symbol='upload')

icon_wizard_step_previous = Icon(driver_name='fontawesome', symbol='caret-left')
icon_wizard_step_first = Icon(driver_name='fontawesome', symbol='step-backward')
icon_wizard_step_next = Icon(driver_name='fontawesome', symbol='caret-right')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Cabinets'), name='cabinets')


event_cabinet_created = namespace.add_event_type(
    label=_('Cabinet created'), name='cabinet_created'
)
event_cabinet_edited = namespace.add_event_type(
    label=_('Cabinet edited'), name='cabinet_edited'
)
event_cabinet_document_added = namespace.add_event_type(
    label=_('Document added to cabinet'), name='add_document'
)
event_cabinet_document_removed = namespace.add_event_type(
    label=_('Document removed from cabinet'), name='remove_document'
)



from django.contrib.auth.views import INTERNAL_RESET_SESSION_TOKEN
from django.core import mail
from django.test import override_settings
from django.urls import reverse

from mayan.apps.common.settings import setting_home_view
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.user_management.events import event_user_edited
from mayan.apps.user_management.permissions import permission_user_edit
from mayan.apps.user_management.tests.literals import (
    TEST_USER_PASSWORD_EDITED
)

from .literals import TEST_PASSWORD_NEW
from .mixins import PasswordResetViewTestMixin, UserPasswordViewTestMixin


class CurrentUserViewTestCase(GenericViewTestCase):
    def _request_current_user_password_change_view(self, new_password):
        return self.post(
            viewname='authentication:password_change_view', data={
                'old_password': self._test_case_user.cleartext_password,
                'new_password1': new_password,
                'new_password2': new_password
            }
        )

    def test_current_user_set_password_view(self):
        self._clear_events()

        response = self._request_current_user_password_change_view(
            new_password=TEST_PASSWORD_NEW
        )
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()
        self.assertTrue(
            self._test_case_user.check_password(
                raw_password=TEST_PASSWORD_NEW
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class PasswordResetViewTestCase(
    PasswordResetViewTestMixin, GenericViewTestCase
):
    auto_login_user = False
    create_test_case_superuser = True

    def test_password_reset_view(self):
        self.logout()
        self._clear_events()

        response = self._request_password_reset_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(len(mail.outbox), 1)
        email_parts = mail.outbox[0].body.replace('\n', '').split('/')
        uidb64 = email_parts[-3]
        token = email_parts[-2]

        # Add the token to the session.
        session = self.client.session
        session[INTERNAL_RESET_SESSION_TOKEN] = token
        session.save()

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

        self._clear_events()

        response = self._request_password_reset_confirm_view(
            new_password=TEST_PASSWORD_NEW, uidb64=uidb64
        )
        self.assertNotIn(INTERNAL_RESET_SESSION_TOKEN, self.client.session)

        self._test_case_superuser.refresh_from_db()
        self.assertTrue(
            self._test_case_superuser.check_password(
                raw_password=TEST_PASSWORD_NEW
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_superuser)
        self.assertEqual(events[0].target, self._test_case_superuser)
        self.assertEqual(events[0].verb, event_user_edited.id)

    @override_settings(AUTHENTICATION_DISABLE_PASSWORD_RESET=False)
    def test_password_reset_get_view_with_disable_false(self):
        self.logout()

        self._clear_events()

        response = self._request_password_reset_get_view()

        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @override_settings(AUTHENTICATION_DISABLE_PASSWORD_RESET=True)
    def test_password_reset_get_view_with_disable_true(self):
        self.logout()

        self._clear_events()

        response = self._request_password_reset_get_view()

        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse(viewname=setting_home_view.value)
        )

        self.assertEqual(len(mail.outbox), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @override_settings(AUTHENTICATION_DISABLE_PASSWORD_RESET=False)
    def test_password_reset_post_view_with_disable_false(self):
        self.logout()

        self._clear_events()

        response = self._request_password_reset_post_view()

        self.assertEqual(response.status_code, 302)
        self.assertNotEqual(response.url, setting_home_view.value)

        self.assertEqual(len(mail.outbox), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @override_settings(AUTHENTICATION_DISABLE_PASSWORD_RESET=True)
    def test_password_reset_post_view_with_disable_true(self):
        self.logout()

        self._clear_events()

        response = self._request_password_reset_post_view()

        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse(viewname=setting_home_view.value)
        )

        self.assertEqual(len(mail.outbox), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class UserPasswordViewTestCase(
    UserPasswordViewTestMixin, GenericViewTestCase
):
    def test_user_set_password_view_no_permission(self):
        self._create_test_user()

        password_hash = self.test_user.password

        self._clear_events()

        response = self._request_test_user_password_set_view(
            password=TEST_USER_PASSWORD_EDITED
        )
        self.assertEqual(response.status_code, 404)

        self.test_user.refresh_from_db()
        self.assertEqual(self.test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_set_password_view_with_access(self):
        self._create_test_user()
        self.grant_access(obj=self.test_user, permission=permission_user_edit)

        password_hash = self.test_user.password

        self._clear_events()

        response = self._request_test_user_password_set_view(
            password=TEST_USER_PASSWORD_EDITED
        )
        self.assertEqual(response.status_code, 302)

        self.test_user.refresh_from_db()
        self.assertNotEqual(self.test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_user_multiple_set_password_view_no_permission(self):
        self._create_test_user()
        password_hash = self.test_user.password

        self._clear_events()

        response = self._request_test_user_password_set_multiple_view(
            password=TEST_USER_PASSWORD_EDITED
        )
        self.assertEqual(response.status_code, 404)

        self.test_user.refresh_from_db()
        self.assertEqual(self.test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_multiple_set_password_view_with_access(self):
        self._create_test_user()
        self.grant_access(obj=self.test_user, permission=permission_user_edit)

        password_hash = self.test_user.password

        self._clear_events()

        response = self._request_test_user_password_set_multiple_view(
            password=TEST_USER_PASSWORD_EDITED
        )
        self.assertEqual(response.status_code, 302)

        self.test_user.refresh_from_db()
        self.assertNotEqual(self.test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)



from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from mayan.apps.databases.classes import BaseBackend

from .settings import (
    setting_authentication_backend, setting_authentication_backend_arguments
)


class AuthenticationBackend(BaseBackend):
    _loader_module_name = 'authentication_backends'
    form_list = ()
    login_form_class = None
    login_form_class_path = 'django.contrib.auth.forms.AuthenticationForm'

    @classmethod
    def cls_get_instance(cls):
        authentication_backend_class = cls.get(
            name=setting_authentication_backend.value
        )
        return authentication_backend_class(
            **setting_authentication_backend_arguments.value
        )

    @classmethod
    def cls_initialize(cls):
        backend = cls.cls_get_instance()
        backend.initialize()

    def initialize(self):
        """
        Optional subclass initialization method.
        """

    def get_condition_dict(self):
        result = {}

        for form_index, form in enumerate(iterable=self.form_list):
            condition = getattr(form, 'condition', None)
            if condition:
                def condition_wrapper(authentication_backend):
                    # SessionWizard form condition method is hard coded
                    # to receive only the wizard instance. This wrapper
                    # allows passing also the `authentication_backend`.
                    def wrapper(wizard):
                        return condition(
                            authentication_backend=authentication_backend,
                            wizard=wizard
                        )

                    return wrapper

                result[str(form_index)] = condition_wrapper(
                    authentication_backend=self
                )

        return result

    def get_form_list(self):
        return self.form_list

    def get_login_form_class(self):
        if not self.login_form_class:
            if not self.login_form_class_path:
                raise ImproperlyConfigured(
                    'Must specify a `login_form_class` or a '
                    '`login_form_class_path`.'
                )
            else:
                return import_string(
                    dotted_path=self.login_form_class_path
                )
        else:
            return self.login_form_class

    def get_user(self, request, form_list=None, kwargs=None):
        """
        Required method to identify the user based on form and request data.
        """
        raise NotImplementedError

    def process(request, form_list=None, kwargs=None):
        """
        Optional method to do login related setup based on form data like
        session TTL.
        """


class AuthenticationBackendRememberMeMixin:
    def __init__(self, **kwargs):
        self.maximum_session_length = kwargs.pop('maximum_session_length')
        super().__init__(**kwargs)

    def process(self, form_list=None, kwargs=None, request=None):
        kwargs = kwargs or {}
        remember_me = kwargs.get('remember_me')

        # remember_me values:
        # True - long session
        # False - short session
        # None - Form has no remember_me value and we let the session
        # expiration default.

        if remember_me is True:
            request.session.set_expiry(
                self.maximum_session_length
            )
        elif remember_me is False:
            request.session.set_expiry(0)



from django.utils.translation import ugettext_lazy as _

INCLUSION_AND = '&'
INCLUSION_OR = '|'

INCLUSION_CHOICES = (
    (INCLUSION_AND, _('and')),
    (INCLUSION_OR, _('or')),
)

OPERATOR_CHOICES = (
    ('exact', _('is equal to')),
    ('iexact', _('is equal to (case insensitive)')),
    ('contains', _('contains')),
    ('icontains', _('contains (case insensitive)')),
    ('in', _('is in')),
    ('gt', _('is greater than')),
    ('gte', _('is greater than or equal to')),
    ('lt', _('is less than')),
    ('lte', _('is less than or equal to')),
    ('startswith', _('starts with')),
    ('istartswith', _('starts with (case insensitive)')),
    ('endswith', _('ends with')),
    ('iendswith', _('ends with (case insensitive)')),
    ('regex', _('is in regular expression')),
    ('iregex', _('is in regular expression (case insensitive)')),
)



from .events import event_parsing_document_file_submitted
from .tasks import task_parse_document_file


def method_document_parsing_submit(self, _user=None):
    latest_file = self.file_latest
    # Don't error out if document has no file.
    if latest_file:
        latest_file.submit_for_parsing(_user=_user)


def method_document_file_parsing_submit(self, _user=None):
    event_parsing_document_file_submitted.commit(
        action_object=self.document, actor=_user, target=self
    )

    if _user:
        user_id = _user.pk
    else:
        user_id = None

    task_parse_document_file.apply_async(
        kwargs={'document_file_pk': self.pk, 'user_id': user_id}
    )



from mayan.apps.documents.tests.base import BaseTestCase

from .mixins import AssetTaskTestMixin, AssetTestMixin


class AssetTasksTestCase(AssetTaskTestMixin, AssetTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_asset()

    def test_asset_task_content_object_image_generate(self):
        self._execute_asset_task_content_object_image_generate()



import logging
from pathlib import Path
import re

from django.core.files import File
from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.models import SharedUploadedFile

from ..classes import SourceBackend
from ..exceptions import SourceException

from .literals import (
    REGULAR_EXPRESSION_MATCH_EVERYTHING, REGULAR_EXPRESSION_MATCH_NOTHING,
    SOURCE_INTERVAL_UNCOMPRESS_CHOICES
)
from .mixins import (
    SourceBackendCompressedMixin, SourceBackendPeriodicMixin, SourceBaseMixin
)

__all__ = ('SourceBackendWatchFolder',)
logger = logging.getLogger(name=__name__)


class SourceBackendWatchFolder(
    SourceBackendCompressedMixin, SourceBackendPeriodicMixin,
    SourceBaseMixin, SourceBackend
):
    field_order = (
        'folder_path', 'include_subdirectories', 'include_regex',
        'exclude_regex'
    )
    fields = {
        'folder_path': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Server side filesystem path.'
            ),
            'kwargs': {
                'max_length': 255,
            },
            'label': _('Folder path'),
            'required': True
        },
        'include_subdirectories': {
            'class': 'django.forms.BooleanField',
            'default': '',
            'help_text': _(
                'If checked, not only will the folder path be scanned for '
                'files but also its subdirectories.'
            ),
            'label': _('Include subdirectories?'),
            'required': False
        },
        'include_regex': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Regular expression used to select which files to upload.'
            ),
            'label': _('Include regular expression'),
            'required': False
        },
        'exclude_regex': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Regular expression used to exclude which files to upload.'
            ),
            'label': _('Exclude regular expression'),
            'required': False
        }
    }
    label = _('Watch folder')
    uncompress_choices = SOURCE_INTERVAL_UNCOMPRESS_CHOICES

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)
        include_regex = re.compile(
            pattern=self.kwargs.get(
                'include_regex', REGULAR_EXPRESSION_MATCH_EVERYTHING
            )
        )
        exclude_regex = re.compile(
            pattern=self.kwargs.get(
                'exclude_regex', REGULAR_EXPRESSION_MATCH_NOTHING
            ) or REGULAR_EXPRESSION_MATCH_NOTHING
        )
        path = Path(self.kwargs['folder_path'])

        # Force testing the path and raise errors for the log.
        path.lstat()
        if not path.is_dir():
            raise SourceException('Path {} is not a directory.'.format(path))

        if self.kwargs['include_subdirectories']:
            iterator = path.rglob(pattern='*')
        else:
            iterator = path.glob(pattern='*')

        for entry in iterator:
            if entry.is_file() or entry.is_symlink():
                if include_regex.match(string=entry.name) and not exclude_regex.match(string=entry.name):
                    with entry.open(mode='rb+') as file_object:
                        shared_uploaded_file = SharedUploadedFile.objects.create(
                            file=File(file=file_object), filename=entry.name
                        )
                        if not dry_run:
                            entry.unlink()

                        return (shared_uploaded_file,)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_d

queue_tools = CeleryQueue(label=_('Tools'), name='tools', worker=worker_d)



from django.db import migrations, models


def operation_remove_duplicates(apps, schema_editor):
    StoredDriver = apps.get_model(
        app_label='file_metadata', model_name='StoredDriver'
    )
    DocumentVersionDriverEntry = apps.get_model(
        app_label='file_metadata', model_name='DocumentVersionDriverEntry'
    )

    # This logic assumes only one type of file metadata backend exists at
    # this point of time.
    driver = StoredDriver.objects.first()
    if driver:
        DocumentVersionDriverEntry.objects.using(alias=schema_editor.connection.alias).update(driver=driver)

        StoredDriver.objects.exclude(pk=driver.id).delete()


class Migration(migrations.Migration):
    atomic = False
    dependencies = [
        ('file_metadata', '0002_documenttypesettings')
    ]
    operations = [
        migrations.RunPython(
            code=operation_remove_duplicates,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AlterField(
            model_name='storeddriver',
            name='driver_path',
            field=models.CharField(
                max_length=255, unique=True, verbose_name='Driver path'
            )
        ),
        migrations.AlterField(
            model_name='storeddriver',
            name='internal_name',
            field=models.CharField(
                db_index=True, max_length=128, unique=True,
                verbose_name='Internal name'
            )
        )
    ]
    run_before = [
        ('documents', '0057_auto_20200916_1057')
    ]



API_VERSION = '4'
DEFAULT_PAGE_SIZE_QUERY_PARAMETER = 'page_size'
DEFAULT_REST_API_DISABLE_LINKS = False
DEFAULT_REST_API_MAXIMUM_PAGE_SIZE = 100
DEFAULT_REST_API_PAGE_SIZE = 10

QUERY_FIELD_EXCLUDE_PARAMETER = '_fields_exclude'
QUERY_FIELD_ONLY_PARAMETER = '_fields_only'



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('linking', '0008_auto_20190429_1922'),
    ]

    operations = [
        migrations.AlterField(
            model_name='smartlink',
            name='dynamic_label',
            field=models.CharField(
                blank=True, help_text='Use this field to show a unique '
                'label depending on the document from which the smart '
                'link is being accessed.', max_length=96,
                verbose_name='Dynamic label'
            ),
        ),
        migrations.AlterField(
            model_name='smartlinkcondition',
            name='expression',
            field=models.TextField(
                help_text='The expression using document properties to '
                'be evaluated against the foreign document field.',
                verbose_name='Expression'
            ),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0017_auto_20191213_0044'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='workflowtransitionfield',
            options={
                'verbose_name': 'Workflow transition field',
                'verbose_name_plural': 'Workflow transition fields'
            },
        ),
    ]



import re

import django.core.validators
from django.db import migrations, models

import mayan.apps.converter.models
import mayan.apps.storage.classes


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0017_auto_20200810_0504'),
    ]

    operations = [
        migrations.CreateModel(
            name='Asset',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        max_length=96, unique=True, verbose_name='Label'
                    )
                ),
                (
                    'internal_name', models.CharField(
                        db_index=True, help_text='This value will be '
                        'used when referencing this asset. Can only '
                        'contain letters, numbers, and underscores.',
                        max_length=255, unique=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile('^[a-zA-Z0-9_]+\\Z'),
                                "Enter a valid 'internal name' consisting "
                                "of letters, numbers, and underscores.",
                                'invalid'
                            )
                        ], verbose_name='Internal name')
                ),
                (
                    'file', models.FileField(
                        storage=mayan.apps.storage.classes.DefinedStorageLazy(
                            name='converter__assets'
                        ), upload_to=mayan.apps.converter.models.upload_to,
                        verbose_name='File'
                    )
                ),
            ],
            options={
                'verbose_name': 'Asset',
                'verbose_name_plural': 'Assets',
                'ordering': ('label',),
            },
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_signatures', '0005_auto_20160325_0748'),
    ]

    operations = [
        migrations.AlterField(
            model_name='signaturebasemodel',
            name='public_key_fingerprint',
            field=models.CharField(
                verbose_name='Public key fingerprint', max_length=40,
                null=True, editable=False, blank=True
            ),
        ),
    ]



from django.utils.encoding import force_text

from mayan.apps.events.classes import EventModelRegistry
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_acl_created, event_acl_deleted, event_acl_edited
from ..models import AccessControlList
from ..permissions import permission_acl_edit, permission_acl_view

from .mixins import ACLTestMixin, AccessControlListViewTestMixin


class AccessControlListViewTestCase(
    AccessControlListViewTestMixin, ACLTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_acl_test_object()

    def test_acl_create_get_view_no_permission(self):
        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_create_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(AccessControlList.objects.count(), acl_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_create_get_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )
        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_create_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(AccessControlList.objects.count(), acl_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_create_view_post_no_permission(self):
        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_create_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(AccessControlList.objects.count(), acl_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_create_view_post_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )

        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_create_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(AccessControlList.objects.count(), acl_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, self.test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_acl)
        self.assertEqual(events[0].verb, event_acl_created.id)

    def test_acl_delete_view_no_permission(self):
        self._create_test_acl()

        self._clear_events()

        response = self._request_test_acl_delete_view()
        self.assertNotContains(
            response=response, text=force_text(s=self.test_object),
            status_code=404
        )

        self.assertTrue(
            self.test_object.acls.filter(role=self.test_role).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_delete_view_with_access(self):
        self._create_test_acl()

        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )

        self._clear_events()

        response = self._request_test_acl_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self.test_object.acls.filter(role=self.test_role).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_object)
        self.assertEqual(events[0].verb, event_acl_deleted.id)

    def test_acl_list_view_no_permission(self):
        self._create_test_acl()

        self._clear_events()

        response = self._request_test_acl_list_view()
        self.assertNotContains(
            response=response, text=force_text(s=self.test_object),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_list_view_with_access(self):
        self._create_test_acl()

        self.grant_access(
            obj=self.test_object, permission=permission_acl_view
        )

        self._clear_events()

        response = self._request_test_acl_list_view()
        self.assertContains(
            response=response, text=force_text(s=self.test_object),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class AccessControlListPermissionViewTestCase(
    AccessControlListViewTestMixin, ACLTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_acl_test_object()
        self._create_test_acl()

    def test_acl_permission_get_no_permission(self):
        test_acl_permission_count = self.test_acl.permissions.count()

        self._clear_events()

        response = self._request_test_acl_permission_list_get_view()
        self.assertNotContains(
            response=response, text=force_text(s=self.test_object),
            status_code=404
        )

        self.assertEqual(
            self.test_acl.permissions.count(), test_acl_permission_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_get_with_access(self):
        test_acl_permission_count = self.test_acl.permissions.count()

        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )

        self._clear_events()

        response = self._request_test_acl_permission_list_get_view()
        self.assertContains(
            response=response, text=force_text(s=self.test_object),
            status_code=200
        )

        self.assertEqual(
            self.test_acl.permissions.count(), test_acl_permission_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_add_view_with_access(self):
        test_acl_permission_count = self.test_acl.permissions.count()

        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )
        test_acl_permission_count = self.test_acl.permissions.count()

        self._clear_events()

        response = self._request_test_acl_permission_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_acl.permissions.count(), test_acl_permission_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_acl)
        self.assertEqual(events[0].verb, event_acl_edited.id)

    def test_acl_permission_remove_view_with_access(self):
        self.test_acl.permissions.add(self.test_permission.stored_permission)

        test_acl_permission_count = self.test_acl.permissions.count()

        self.grant_access(
            obj=self.test_object, permission=permission_acl_edit
        )
        test_acl_permission_count = self.test_acl.permissions.count()

        self._clear_events()

        response = self._request_test_acl_permission_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_acl.permissions.count(), test_acl_permission_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_acl)
        self.assertEqual(events[0].verb, event_acl_edited.id)


class GlobalAccessControlListViewTestCase(
    AccessControlListViewTestMixin, ACLTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_acl_test_object()
        self._create_test_acl()

    def test_global_acl_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_global_acl_list_view()
        self.assertNotContains(
            response=response, text=force_text(s=self.test_object),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_global_acl_list_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_acl_view
        )

        self._clear_events()

        response = self._request_test_global_acl_list_view()
        self.assertContains(
            response=response, text=force_text(s=self.test_object),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class OrphanAccessControlListViewTestCase(
    AccessControlListViewTestMixin, ACLTestMixin, GenericViewTestCase
):
    def test_orphan_acl_create_view_with_permission(self):
        """
        Test creating an ACL entry for an object with no model permissions.
        Result: Should display a blank permissions list (no optgroup)
        """
        self._create_acl_test_object()
        EventModelRegistry.register(model=self.TestModel)

        self.grant_permission(permission=permission_acl_edit)

        test_acl_count = AccessControlList.objects.count()

        self._clear_events()

        reponse = self._request_test_acl_create_post_view()
        self.assertEqual(reponse.status_code, 302)

        response = self._request_test_acl_create_get_view()
        self.assertNotContains(
            response=response, text='optgroup', status_code=200
        )

        self.assertEqual(
            AccessControlList.objects.count(), test_acl_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_acl)
        self.assertEqual(events[0].verb, event_acl_created.id)



from mayan.apps.appearance.classes import Icon

icon_otp_detail = Icon(driver_name='fontawesome', symbol='shield-alt')
icon_otp_disable = Icon(
    driver_name='fontawesome-dual', primary_symbol='shield-alt',
    secondary_symbol='minus'
)
icon_otp_enable = Icon(
    driver_name='fontawesome-dual', primary_symbol='shield-alt',
    secondary_symbol='plus'
)



from collections import Iterable

from mayan.apps.common.utils import (
    ResolverPipelineModelAttribute, flatten_list
)

from .classes import SearchBackend
from .tasks import task_deindex_instance, task_index_instance


def handler_deindex_instance(sender, **kwargs):
    instance = kwargs['instance']

    task_deindex_instance.apply_async(
        kwargs={
            'app_label': instance._meta.app_label,
            'model_name': instance._meta.model_name,
            'object_id': instance.pk
        }
    )


def handler_factory_index_related_instance_delete(reverse_field_path):
    def handler_index_by_related_to_delete_instance(sender, **kwargs):
        related_instance = kwargs['instance']

        result = ResolverPipelineModelAttribute.resolve(
            attribute=reverse_field_path, obj=related_instance
        )

        entries = flatten_list(value=result)

        def call_task(instance):
            task_index_instance.apply_async(
                kwargs={
                    'app_label': instance._meta.app_label,
                    'model_name': instance._meta.model_name,
                    'object_id': instance.pk,
                    'exclude_app_label': related_instance._meta.app_label,
                    'exclude_model_name': related_instance._meta.model_name,
                    'exclude_kwargs': {'id': related_instance.pk}
                }
            )

        if isinstance(entries, Iterable):
            for instance in entries:
                call_task(instance=instance)
        else:
            call_task(instance=result)

    return handler_index_by_related_to_delete_instance


def handler_factory_index_related_instance_save(reverse_field_path):
    def handler_index_by_related_instance(sender, **kwargs):
        related_instance = kwargs['instance']

        result = ResolverPipelineModelAttribute.resolve(
            attribute=reverse_field_path, obj=related_instance
        )

        entries = flatten_list(value=result)

        def call_task(instance):
            task_index_instance.apply_async(
                kwargs={
                    'app_label': instance._meta.app_label,
                    'model_name': instance._meta.model_name,
                    'object_id': instance.pk
                }
            )

        if isinstance(entries, Iterable):
            for instance in entries:
                call_task(instance=instance)
        else:
            call_task(instance=result)

    return handler_index_by_related_instance


def handler_factory_index_related_instance_m2m(data):
    def handler_index_related_instance_m2m(sender, **kwargs):
        action = kwargs.get('action')
        instance = kwargs['instance']
        model = kwargs.get('model')

        if action in ('post_add', 'pre_remove'):
            instance_paths = data.get(instance._meta.model, ())
            model_paths = data.get(model, ())

            if action == 'pre_remove':
                exclude_kwargs = {
                    'exclude_app_label': instance._meta.app_label,
                    'exclude_model_name': instance._meta.model_name,
                    'exclude_kwargs': {'id': instance.pk}
                }
            else:
                exclude_kwargs = {}

            for instance_path in instance_paths:
                result = ResolverPipelineModelAttribute.resolve(
                    attribute=instance_path, obj=instance
                )

                entries = flatten_list(value=result)

                for entry in entries:
                    task_kwargs = {
                        'app_label': entry._meta.app_label,
                        'model_name': entry._meta.model_name,
                        'object_id': entry.pk
                    }
                    task_kwargs.update(exclude_kwargs)

                    task_index_instance.apply_async(
                        kwargs=task_kwargs
                    )

            if action == 'pre_remove':
                exclude_kwargs = {
                    'exclude_app_label': model._meta.app_label,
                    'exclude_model_name': model._meta.model_name,
                    'exclude_kwargs': {'id__in': list(kwargs['pk_set'])}
                }
            else:
                exclude_kwargs = {}

            for model_instance in model._meta.default_manager.filter(pk__in=kwargs['pk_set']):
                for instance_path in model_paths:
                    result = ResolverPipelineModelAttribute.resolve(
                        attribute=instance_path, obj=model_instance
                    )

                    entries = flatten_list(value=result)

                    for entry in entries:
                        task_kwargs = {
                            'app_label': entry._meta.app_label,
                            'model_name': entry._meta.model_name,
                            'object_id': entry.pk
                        }
                        task_kwargs.update(exclude_kwargs)

                        task_index_instance.apply_async(
                            kwargs=task_kwargs
                        )

    return handler_index_related_instance_m2m


def handler_index_instance(sender, **kwargs):
    instance = kwargs['instance']

    task_index_instance.apply_async(
        kwargs={
            'app_label': instance._meta.app_label,
            'model_name': instance._meta.model_name,
            'object_id': instance.pk
        }
    )


def handler_search_backend_initialize(sender, **kwargs):
    backend = SearchBackend.get_instance()

    backend.initialize()


def handler_search_backend_upgrade(sender, **kwargs):
    backend = SearchBackend.get_instance()

    backend.upgrade()



from django.db import migrations, models
import django.db.models.deletion

DEFAULT_FALSE_CONTENT_TYPE_ID = 1
DEFAULT_FALSE_OBJECT_ID = 1


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('logging', '0003_auto_20211201_1055')
    ]

    operations = [
        migrations.CreateModel(
            bases=('logging.errorlogpartitionentry',), fields=[],
            name='GlobalErrorLogPartitionEntry', options={
                'constraints': [],
                'indexes': [],
                'proxy': True
            }
        ),
        migrations.AddField(
            field=models.ForeignKey(
                default=DEFAULT_FALSE_CONTENT_TYPE_ID,
                on_delete=django.db.models.deletion.CASCADE,
                to='contenttypes.contenttype'
            ), model_name='errorlogpartition', name='content_type',
            preserve_default=False
        ),
        migrations.AddField(
            field=models.PositiveIntegerField(
                default=DEFAULT_FALSE_OBJECT_ID
            ), model_name='errorlogpartition', name='object_id',
            preserve_default=False
        ),
    ]



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import document_search
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.tags.tests.mixins import TagTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import DynamicSearchException

from .mixins import SearchTestMixin


class QueryStringDecodeTestCase(SearchTestMixin, BaseTestCase):
    def test_decode_default_scope(self):
        query = {
            'test_field': 'test_value'
        }

        self.assertEqual(
            self.search_backend.decode_query(query=query), {
                'operators': {}, 'result_scope': '0',
                'scopes': {
                    '0': {
                        'match_all': False, 'query': {
                            'test_field': 'test_value'
                        }
                    }
                }
            }
        )

    def test_decode_default_scope_with_explicit_scope(self):
        query = {
            'test_field': 'test_value',
            '__1_test_field': 'test_value',
        }

        self.assertEqual(
            self.search_backend.decode_query(query=query), {
                'operators': {}, 'result_scope': '0',
                'scopes': {
                    '0': {
                        'match_all': False, 'query': {
                            'test_field': 'test_value'
                        }
                    },
                    '1': {
                        'match_all': False, 'query': {
                            'test_field': 'test_value'
                        }
                    }
                }
            }
        )

    def test_decode_scope_0_match_all(self):
        query = {
            '__0_test_field': 'test_value',
            '__0_match_all': 'true',
            '__1_test_field': 'test_value',
        }

        self.assertEqual(
            self.search_backend.decode_query(query=query), {
                'operators': {}, 'result_scope': '0',
                'scopes': {
                    '0': {
                        'match_all': True, 'query': {
                            'test_field': 'test_value'
                        }
                    },
                    '1': {
                        'match_all': False, 'query': {
                            'test_field': 'test_value'
                        }
                    }
                }
            }
        )

    def test_decode_scope_1_match_all(self):
        query = {
            '__0_test_field': 'test_value',
            '__0_match_all': 'false',
            '__1_test_field': 'test_value',
            '__1_match_all': 'true',
        }

        self.assertEqual(
            self.search_backend.decode_query(query=query), {
                'operators': {}, 'result_scope': '0',
                'scopes': {
                    '0': {
                        'match_all': False, 'query': {
                            'test_field': 'test_value'
                        }
                    },
                    '1': {
                        'match_all': True, 'query': {
                            'test_field': 'test_value'
                        }
                    }
                }
            }
        )


class ScopedSearchTestCase(
    DocumentTestMixin, SearchTestMixin, TagTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()
        self._create_test_tag()
        self._create_test_tag()

        self.test_tags[0].documents.add(self.test_document)
        self.test_tags[1].documents.add(self.test_document)

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._index_instance(instance=self.test_document)
        self._index_instance(instance=self.test_tags[0])
        self._index_instance(instance=self.test_tags[1])

    def test_missing_scope_query(self):
        query = {
            '__a_match_all': 'true',
            '__operator_a_b': 'OR_c',
            '__b_label': self.test_documents[0].label,
            '__result': 'c'
        }
        with self.assertRaises(expected_exception=DynamicSearchException):
            self.search_backend.search(
                search_model=document_search, query=query,
                user=self._test_case_user
            )

    def test_AND_scope(self):
        query = {
            '__0_tags__label': self.test_tags[0].label,
            '__operator_0_1': 'AND_901',
            '__result': '901',
            '__1_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

        self.test_tags[1].documents.remove(self.test_document)
        self._index_instance(instance=self.test_document)
        self._index_instance(instance=self.test_tags[1])

        query = {
            '__0_tags__label': self.test_tags[0].label,
            '__operator_0_1': 'AND_901',
            '__result': '901',
            '__1_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_OR_scope(self):
        query = {
            '__0_tags__label': self.test_tags[0].label,
            '__operator_0_1': 'OR_901',
            '__result': '901',
            '__1_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

    def test_letter_scopes(self):
        query = {
            '__a_tags__label': self.test_tags[0].label,
            '__operator_a_b': 'OR_c',
            '__result': 'c',
            '__b_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

    def test_multi_letter_scopes(self):
        query = {
            '__ab_tags__label': self.test_tags[0].label,
            '__operator_ab_bc': 'OR_cc',
            '__result': 'cc',
            '__bc_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

    def test_single_scope(self):
        query = {
            '__ab_tags__label': self.test_tags[0].label,
            '__operator_ab_bc': 'OR_cc',
            '__result': 'ab',
            '__bc_tags__label': self.test_tags[1].label
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)
        self.assertTrue(self.test_documents[0] in queryset)

    def test_match_all_scope_and_non_match_all_scope(self):
        query = {
            '__0_match_all': 'TRUE',
            '__0_tags__label': self.test_tags[0].label,
            '__0_tags__color': self.test_tags[0].color,
            '__operator_0_bc': 'AND_cc',
            '__bc_tags__label': self.test_tags[1].label,
            '__bc_tags__color': 'INVALID COLOR',
            '__result': 'cc'
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)
        self.assertTrue(self.test_documents[0] in queryset)


class ScopeOperatorSearchTestCase(
    DocumentTestMixin, SearchTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()
        self._create_test_document_stub()

        self.grant_access(
            obj=self.test_documents[0], permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )

        self._index_instance(instance=self.test_documents[0])
        self._index_instance(instance=self.test_documents[1])

    def test_and_operator_both_scopes_with_data(self):
        query = {
            '__0_label': self.test_documents[0].label,
            '__operator_0_1': 'OR_2',
            '__1_label': self.test_documents[1].label,
            '__result': '2'
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 2)

    def test_and_operator_scope_1_with_data(self):
        query = {
            '__0_label': self.test_documents[0].label,
            '__operator_0_1': 'AND_2',
            '__1_label': 'invalid',
            '__result': '2'
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_and_operator_scope_2_with_data(self):
        query = {
            '__0_label': 'invalid',
            '__operator_0_1': 'AND_2',
            '__1_label': self.test_documents[1].label,
            '__result': '2'
        }
        queryset = self.search_backend.search(
            search_model=document_search, query=query,
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)



from django.template import Library

from ..classes import SearchModel
from ..literals import SEARCH_MODEL_NAME_KWARG

register = Library()


@register.simple_tag
def search_get_default_search_model():
    return SearchModel.get_default()


@register.simple_tag
def search_get_search_models():
    return SearchModel.all()


@register.simple_tag
def search_get_search_model_name_query_variable():
    return '_{}'.format(SEARCH_MODEL_NAME_KWARG)



from .tasks import task_duplicates_clean_empty_lists, task_duplicates_scan_for


def handler_scan_duplicates_for(sender, instance, **kwargs):
    task_duplicates_scan_for.apply_async(
        kwargs={'document_id': instance.document_id}
    )


def handler_remove_empty_duplicates_lists(sender, **kwargs):
    task_duplicates_clean_empty_lists.apply_async()



from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.serializer_mixins import CreateOnlyFieldSerializerMixin
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField
from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.views.utils import resolve

from .models import Message


class MessageSerializer(
    CreateOnlyFieldSerializerMixin, serializers.HyperlinkedModelSerializer
):
    sender_app_label = serializers.SerializerMethodField()
    sender_model_name = serializers.SerializerMethodField()
    sender_url = serializers.SerializerMethodField()
    user = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the recipient user of this message.'
        ), source_queryset=get_user_queryset(),
    )

    class Meta:
        create_only_fields = ('user',)
        extra_kwargs = {
            'url': {
                'lookup_url_kwarg': 'message_id',
                'view_name': 'rest_api:message-detail'
            }
        }
        fields = (
            'body', 'date_time', 'id', 'read', 'sender_app_label',
            'sender_model_name', 'sender_object_id', 'sender_url', 'subject',
            'url', 'user'
        )
        model = Message
        read_only_fields = (
            'date_time', 'id', 'sender_app_label', 'sender_model_name',
            'sender_object_id', 'sender_url', 'url'
        )

    def get_fields(self, *args, **kwargs):
        fields = super().get_fields(*args, **kwargs)
        request = self.context['request']

        if request.method in ('PATCH', 'PUT'):
            fields['body'].read_only = True
            fields['subject'].read_only = True
        elif request.method in ('POST',):
            fields.pop('read')

        return fields

    def get_sender_app_label(self, instance):
        sender_content_type = getattr(instance, 'sender_content_type', None)

        if sender_content_type:
            return sender_content_type.app_label

    def get_sender_model_name(self, instance):
        sender_content_type = getattr(instance, 'sender_content_type', None)

        if sender_content_type:
            return sender_content_type.model

    def get_sender_url(self, instance):
        sender_object = getattr(instance, 'sender_object', None)

        if sender_object:
            try:
                path = sender_object.get_absolute_api_url()
            except AttributeError:
                return
            else:
                if path:
                    resolved_match = resolve(path=path)

                    return reverse(
                        viewname=resolved_match.view_name,
                        kwargs=resolved_match.kwargs, request=self.context['request'],
                        format=self.context['format']
                    )



from rest_framework import status

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..classes import SearchModel

from .mixins import SearchAPIViewTestMixin


class SearchModelAPIViewTestCase(BaseAPITestCase):
    def test_search_models_api_view(self):
        response = self.get(
            viewname='rest_api:searchmodel-list', query={'page_size': 50}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            [search_model['pk'] for search_model in response.data['results']],
            [search_model.pk for search_model in SearchModel.all()]
        )


class SearchAPIViewTestCase(
    DocumentTestMixin, SearchAPIViewTestMixin, BaseAPITestCase
):
    def test_search_api_view_no_permission(self):
        response = self._request_search_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

    def test_search_api_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_search_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'], self.test_document.label
        )
        self.assertEqual(response.data['count'], 1)

    def test_advanced_search_api_view_no_permission(self):
        response = self._request_advanced_search_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

    def test_advanced_search_api_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_advanced_search_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'], self.test_document.label
        )
        self.assertEqual(response.data['count'], 1)



from django.utils.module_loading import import_string

from mayan.apps.storage.utils import TemporaryDirectory

from ..classes import SearchBackend
from ..backends.elasticsearch import ElasticSearchBackend
from ..backends.whoosh import WhooshSearchBackend
from ..literals import DEFAULT_SEARCH_BACKEND
from ..settings import setting_backend_arguments


class TestSearchBackend(SearchBackend):
    """
    Test search backend dynamically generated from the default search
    backend. If the default backend is the Whoosh backend, a temporary
    folder will be created to store the search index. The temporary directory
    is automatically deleted by Python when all references are removed.

    This test search backend will only perform instance indexing and
    deindexing when called from a search test. This allows using the backend
    for search tests but avoids the indexing penalty for non search tests.

    It will also automatically insert the test object into the index
    before performing a call to the `search` method.
    """
    _test_backend_initialized = False
    _test_backend_path = None
    _test_class = None

    def __init__(self, *args, **kwargs):
        if self._test_class:
            backend_path = getattr(self._test_class, '_test_search_backend_path', DEFAULT_SEARCH_BACKEND)
        else:
            backend_path = DEFAULT_SEARCH_BACKEND
            SearchBackend._disable()

        if self.__class__._test_backend_path != backend_path:
            self.__class__._test_backend_path = backend_path

            self.__class__._backend_kwargs = setting_backend_arguments.value.copy()
            self.__class__._backend_class = import_string(dotted_path=backend_path)

            if issubclass(self._backend_class, WhooshSearchBackend):
                if not hasattr(self.__class__, '_local_attribute_backend_temporary_directory'):
                    self.__class__._local_attribute_backend_temporary_directory = TemporaryDirectory()

                self.__class__._backend_kwargs['index_path'] = self._local_attribute_backend_temporary_directory.name
            elif issubclass(self._backend_class, ElasticSearchBackend):
                self.__class__._backend_kwargs['indices_namespace'] = 'test'

            self.__class__._test_backend_initialized = False

        self.__class__._backend = self._backend_class(**self.__class__._backend_kwargs)

        if not self.__class__._test_backend_initialized:
            self.__class__._test_backend_initialized = True
            self._backend._initialize()
            SearchBackend._search_field_transformations = {}

        super().__init__(*args, **kwargs)

    def _search(self, *args, **kwargs):
        return self._backend._search(*args, **kwargs)

    def deindex_instance(self, *args, **kwargs):
        if self._test_class:
            return self._backend.deindex_instance(*args, **kwargs)

    def get_status(self, *args, **kwargs):
        return self._backend.get_status(*args, **kwargs)

    def index_instance(self, *args, **kwargs):
        if self._test_class:
            return self._backend.index_instance(*args, **kwargs)

    def index_search_model(self, *args, **kwargs):
        return self._backend.index_search_model(*args, **kwargs)

    def reset(self, *args, **kwargs):
        return self._backend.reset(*args, **kwargs)

    def tear_down(self):
        self.__class__._test_backend_path = None
        return self._backend.tear_down()
